# -*- coding: utf-8 -*-
from __future__ import with_statement # Python 2.5+
import os, sys, cgi, time, re
from datetime import datetime, timedelta
from urlparse import urlparse
from xml.dom.minidom import parseString
from xml.parsers.expat import ExpatError
    
from bikini.app import *
from bikini import storage, config

# -------------------
# Index page
# -------------------

@view()
def index(environ, filler):     
    return page(environ, filler, config.get('wiki', 'index_page'))

# -------------------
# System pages
# -------------------

@view(r'^/RecentChanges$')
def recent_changes(environ, filler):

    get_first = parse_query_string(environ)
    
    try:
        days = int(get_first('days', REVISION_LIFESPAN_DAYS))
    except ValueError:
        days = REVISION_LIFESPAN_DAYS
                    
    conn = storage.connect()    
    changes = storage.get_recent_changes(conn, days)

        
    template_dd = u"""<tr>
        <td class="dim $is_deleted"><a rel="nofollow" title="View source for revision $revision_id" href="$name/source/$revision_id">Rev. $revision_id</a> $diff_link</td>
        <td>$created_on</td>
        <td>$author_link from $host_name</td>
        <td class="revision-message">$revision_message</td>
        </tr>"""            
    template_dt = u'<tr><th colspan="4"><a href="$name">$name</a></th></tr>'

    page_list = []
    prev_page_name = None
    for page in changes:        
        name = page['name']
        created_on = format_datetime(page['created_on'], short=False)        
        host_name = page['host_name']
        revision_id = page['revision_id'] 
        revision_message = page['message'] or ''
        author_link = make_author_link(page['author'])
        diff_link = make_diff_link(name, page['revision_id'])
        is_deleted = 'revision-deleted' if page['is_deleted'] else ''
        
        #@@TODO Use filler_
        
        d = locals()
        
        if name == prev_page_name:                        
            page_list.append(fill_template(template_dd, **d))
        else:
            page_list.append(fill_template(template_dt, **d))
            page_list.append(fill_template(template_dd, **d))
    
        prev_page_name = name           

    #page_title = name = 'RecentChanges'
    page_list = u'\n'.join(page_list)

    if days:
        subhead = fill_template(u'Showing changes within the last $days days', days=days)
    else:        
        subhead = u'Showing all changes'

        
    return HTTP_OK, [], make_page('meta_changes.html', filler, locals())


@view('^/RecentChanges/feed$')
def recent_changes_feed(environ, filler):

    template_feed_entry = u'''<entry>
        <id>tag:$server_name,$created_on_year:$path/$name#$revision_id</id>
        <link rel="alternate" href="$base_uri/$name/diff/$prev_revision_id/$revision_id" type="text/html" />
        <title>$name$is_new</title>
        <updated>$created_on</updated>
        <author>
            <name>$author</name>
        </author>
        <summary>$summary</summary>
    </entry>'''

# @@TODO AtomRevision http://tools.ietf.org/html/draft-snell-atompub-revision-00
#     template_feed_entry = u'''<entry>
#         <id>tag:$server_name,$created_on_year:$path/$name</id>
#         <link rel="alternate" href="$base_uri/$name" type="text/html" />
#         <link rel="history" href="$base_uri/$name/revisions" type="text/html" />        
#         <link rel="diff" href="$base_uri/$name/diff/$prev_revision_id/$revision_id" type="text/html" />
#         <title>$name$is_new</title>
#         <updated>$created_on</updated>
#         <author>
#             <name>$author</name>
#             <ar:host>$host_name</ar:host>
#         </author>
#         <summary>
#           $summary
#         </summary>
#         <ar:revision
#             number="$revision_id"
#             when="$created_on"
#             initial="$is_initial"
#             type="text">$revision_message</ar:revision>
#     </entry>'''
    
    days = min(REVISION_LIFESPAN_DAYS, 30)
    
    server_name = environ['SERVER_NAME']    
    base_uri = get_base_uri(environ)
    path = urlparse(base_uri).path
    
    conn = storage.connect()    
    changes = storage.get_recent_changes(conn, days)

    # Fallback value
    last_updated_on = datetime.utcnow()

    entry_list = []
    for index, revision in enumerate(changes):        
        # Most recent revision is on top
        if index == 0:
            last_updated_on = revision['created_on']        

        name = revision['name']
        created_on = format_iso_datetime(revision['created_on'])        
        host_name = revision['host_name']
        revision_id = revision['revision_id'] 
        author = revision['author']
        created_on_year = revision['created_on'].year
        revision_message = revision['message'] or ''
        is_new = (revision_id == 1) and u' (New page)' or u''                        
        if revision_id == 1:
            is_initial = 'yes'
            summary = fill_template(u'A new page ‘$name’ has been created.', **locals())
        else:
            is_initial = 'no'        
            prev_revision_id = revision_id-1
            summary = fill_template(u'Revision $revision_id for page ‘$name’ has been created.', **locals())
        
        entry_list.append(fill_template(template_feed_entry, **locals()))
        
    entry_list = u'\n'.join(entry_list)
    last_updated_on = format_iso_datetime(last_updated_on)
        
    return HTTP_OK, [('Content-Type', 'application/atom+xml')], make_page('meta_changes.atom', filler, locals())



def make_page_listing(pages):
    if pages:
        listing = [fill_template(u'<li><a href="$name">$name</a></li>', name=name) for name in pages]
    else:
        listing = [u'<li class="empty">No pages found.</li>']    
        
    return len(pages), u'\n'.join(listing)

    
@view(r'^/Names$')
def names(environ, filler):

    conn = storage.connect()
    
    pages = storage.get_all_names(conn)
    pages.extend(RESERVED_NAMES)
    pages.sort()

    page_count, page_list = make_page_listing(pages)
    
    s = (page_count != 1) and u's' or u''
    page_title = name = 'Names'    
    
    return  HTTP_OK, [], make_page('meta_names.html', filler, locals())


@view(r'^/Needed$')
def needed(environ, filler):

    conn = storage.connect()    

    pages = list(set(storage.get_outbound_links(conn, u'%')) - set(storage.get_all_names(conn)) - set(RESERVED_NAMES))
    pages.sort()
    
    page_count, page_list = make_page_listing(pages)
    
    s = (page_count != 1) and u's' or u''
    #page_title = name = 'Needed'       
    return  HTTP_OK, [], make_page('meta_needed.html', filler, locals())
    

@view(r'^/Unlinked$')
def unlinked(environ, filler):

    conn = storage.connect()  
    
    pages = list(set(storage.get_all_names(conn)) -  set(storage.get_outbound_links(conn, u'%')) - set(RESERVED_NAMES))
    pages.sort()
        
    page_count, page_list = make_page_listing(pages)
    
    page_count = len(pages)
    s = (page_count != 1) and u's' or u''
    #page_title = name = 'Unlinked'
    return HTTP_OK, [], make_page('meta_unlinked.html', filler, locals())   


@view(r'^/AboutBikini$')
def about(environ, filler):
    #page_title = name = 'AboutBikini'
    return  HTTP_OK, [],  make_page('meta_about.html', filler, locals())

@view(r'^/Maintenance$')
def maintenance(environ, filler):
    #page_title = name = 'Maintenance'
    
    button = u'<button type="submit">Run Maintenance Now</button>'
    
    try:
        mtime = datetime.utcfromtimestamp(os.path.getmtime(LOG_PATH))      
        last_run_on = u'Last run on %s UTC.' % format_datetime(mtime, short=False)

        if datetime.utcnow() < (mtime+timedelta(hours=12)):
            button = u'<div class="dim"><i>Cannot run maintenance for the next 12 hours.</i></div>'          
    
    except OSError:
        last_run_on = u'Never run.'
        
    days = REVISION_LIFESPAN_DAYS
    
    message = make_message(get_message(environ))

    cookie = SimpleCookie()
    if message:
      set_message(cookie, '')

    return  HTTP_OK, make_cookie_headers(cookie), make_page('meta_maintenance.html', filler, locals())

@view(r'^/Maintenance$', 'POST')
def maintenance_post(environ, filler):
    #page_title = name = 'Maintenance'

    # Sanity check
    try:
        mtime = datetime.utcfromtimestamp(os.path.getmtime(LOG_PATH))      
        if datetime.utcnow() < (mtime+timedelta(hours=12)):
            #@@NOTE: Perhaps we should use 409 Conflict status code
            return HTTP_FORBIDDEN, [], u'Cannot run maintenance for the next 12 hours.'
            
    except OSError:
        pass
        
    get_first = parse_post_data(environ)
        
    options = get_first('rebuild_pages'), get_first('purge_pages'), get_first('purge_revisions')

    cookie = SimpleCookie()
    
    if any(options):
        hostname = environ.get('REMOTE_HOST', '') or environ.get('REMOTE_ADDR', '')
        user = environ.get('REMOTE_USER', '') or get_author(environ)    
        stats = run_maintenance(user, hostname, *options)    
        set_message(cookie, u'INFO ✓ Maintenance run successully: %d purged pages, %d purged revisions and %d rebuilt pages.' % tuple(stats))
    else:
        set_message(cookie, u'WARNING No action performed, please select maintenance options below.')


    return redirect_to('Maintenance', headers=make_cookie_headers(cookie))

@view(r'^/UserPreferences$')
def preferences(environ, filler):
    author = get_author(environ)
 
    message = make_message(get_message(environ))

    cookie = SimpleCookie()
    if message:
      set_message(cookie, '')
 
    #page_title = name = 'UserPreferences'
    return HTTP_OK, make_cookie_headers(cookie), make_page('meta_preferences.html', filler, locals())

@view(r'^/UserPreferences$', 'POST')
def preferences_post(environ, filler):
    
    cookie = SimpleCookie()
    
    get_first = parse_post_data(environ)    

    set_message(cookie, u'INFO ✓ Your preferences has been saved.')    
    set_author(cookie, get_first('author'))

    return redirect_to('UserPreferences', headers=make_cookie_headers(cookie))


@view(r'^/Macros$')
def macros(environ, filler):
    
    builtin_macros = load_builtin_macros()

    macro_list = []
    for name, function in builtin_macros.items():        
        doc = process_content(function.__doc__, verbatim_macros=True).emit()
        macro_list.append(fill_template(u'<dt>$name</dt>\n<dd>$doc</dd>', name=name.capitalize(), doc=doc))   

    macro_count = len(macro_list)    
    s = (macro_count != 1) and u's' or u''
    
    macro_list = u'\n'.join(macro_list)
    
    return  HTTP_OK, [], make_page('meta_macros.html', filler, locals())

 
    
    
# @view(r'^/Export$')
# def export(environ, filler):
#     #page_title = name = 'UserPreferences'
#     return HTTP_OK, [], make_page('meta_export.html', filler, locals())
#     
# @view(r'^/Export$', 'POST')
# def export_post(environ, filler):
# 
#   #@@TODO: urlquote <link> contents
#   template_feed_entry = u'''<entry>
#       <id>tag:$server_name,$updated_on_year:$path/$name</id>
#       <link href="$base_uri/$name" rel="alternate" type="text/html" />
#       <title>$name</title>
#       <updated>$updated_on</updated>
#       <author>
#           <name>$author</name>
#       </author>
#       <content type="text">
#           $content
#       </content>
#   </entry>'''
# 
#   entry_list = []
# 
#   feed_last_updated_on = format_iso_datetime(datetime.utcnow())
#   base_uri = get_base_uri(environ)
#     
#   server_name = urlparse(base_uri).netloc
#   path = urlparse(base_uri).path
# 
#   conn = storage.connect()  
#   
#   for name in storage.get_all_names(conn):           
#     page = storage.get_page(conn, name)
#     
#     name = page['name']
#     updated_on = format_iso_datetime(page['last_updated_on'])        
#     author = page['last_author']
#     updated_on_year = page['last_updated_on'].year
#     content = escape_html(page['content'])
#     
#     entry_list.append(fill_template(template_feed_entry, **locals()))
#         
# 
#   entry_list = u'\n'.join(entry_list)
#         
# 
#   return HTTP_OK, [('Content-Type', 'application/octet-stream'), ('Content-Disposition', 'attachment; filename="feed.atom"')], make_page('meta_export.atom', filler, locals())

# -------------------
# Show a page
# -------------------

@view(r'^/(%s)$' % WIKI_WORD)
def page(environ, filler, name):

    message = make_message(get_message(environ))
   
    conn = storage.connect()

    page = storage.get_page(conn, name)
    if not page:
        #raise NotFoundError(body=u'Page not found')        
        return HTTP_NOT_FOUND, [], make_page('page_not_found.html', filler, locals())

    last_updated_on = format_datetime(page['last_updated_on'])
    
    content_html = page['content_html']
    revision_id = page['revision_id']
    author = make_author_link(page['last_author'])
    page_title = page['title']
    template_li = u'<a href="$link_id">$link_id</a>'

    inbound_links = []
    
    for link_id in storage.get_inbound_links(conn, name):    
        inbound_links.append(fill_template(template_li, **locals()))

    subhead = u''
    if inbound_links:
        subhead = u'Nearby &rsaquo; %s' % u', '.join(inbound_links)

    diff_link = make_diff_link(name, revision_id)

    cookie = SimpleCookie()
    if message:
      set_message(cookie, '')

    return HTTP_OK, make_cookie_headers(cookie), make_page('page.html', filler, locals())    
    #@@TODO Pass E-tag header too with content_hash
    #return make_page(u'page.html', headers={'Last-Modified':format_http_date(page['last_updated_on'])}, **locals())

# def find_page_title(content):    
# 
#     def get_text(heading_node):
#         # Manage line breaks too
#         return u' '.join([node.nodeValue for node in heading_node.childNodes if node.nodeType == node.TEXT_NODE])
# 
#     try:
#         document_node = parseString(encode(content))
#     except ExpatError:
#         return
#         
#     try:
#         # Grab first
#         h1 = document_node.getElementsByTagName('h1')[0]
#     except IndexError:
#         return
# 
#     return get_text(h1)


# -------------------
# Create/update a page
# -------------------

@view(r'^/(%s)$' % WIKI_WORD, 'POST')
def page_post(environ, filler, name):

    cookie = SimpleCookie()

    get_first = parse_post_data(environ)
          
    content = get_first('content')
    honeypot = get_first('hneoyopt')   
    
    if honeypot:
        # Ignore request 
        set_message(cookie, u'WARNING The changes have been ignored — are you a spambot?')
        return redirect_to(name)     

    content = decode(content)

    is_deleted = False
#@@TODO: check if deleting    
#     deleted_markers = 'PageDeleted', 'DeletedPage'
#     is_deleted = False
#     if content.lstrip().startswith(deleted_markers):
#         is_deleted = True

    emitter = process_content(content, macros=load_builtin_macros())    
    # Wrap everything up within a div -  
    #  this helps further processig with DOM parsers 
    content_html = u'<div>\n%s\n</div>' % emitter.emit()    

    if config.getboolean('appearance', 'smarty_pants'):
        from bikini import smartypants
        content_html = smartypants.smartyPants(content_html, 'qbDe')    

    author = get_author(environ)
    host_name = environ.get('REMOTE_HOST', '') or environ.get('REMOTE_ADDR', '')
    page_title = name

    conn = storage.connect()
    page = storage.get_page(conn, name)
    with conn:
        content_hash = make_hash(content)
        if page: 
            # Check if it's changed something    
            if (page['content_hash'] == content_hash) or (EMPTY_HASH == content_hash):
                set_message(cookie, u'INFO No changes made.')
                return redirect_to(name, headers=make_cookie_headers(cookie))

            new_revision_id = page['revision_id'] + 1
            storage.update_page(conn, name, page_title, content_html, content_hash, is_deleted)
            storage.archive_revision(conn, name, page['revision_id'])
            storage.add_revision(conn, name, new_revision_id, content, author, host_name, message='Page deleted.' if is_deleted else '')    
            if is_deleted:
                set_message(cookie, u'INFO ✓ Page marked for deletion.') 
            else:
                set_message(cookie, u'INFO ✓ Changes saved to revision %d.' % new_revision_id) 
        else:        
            storage.add_page(conn, name, page_title, content_html, content_hash)
            storage.add_revision(conn, name, 1, content, author, host_name, message=u'Page created.')    
            set_message(cookie, u'INFO ✓ Page created.')             
    
        storage.update_links(conn, name, emitter.outbound_links)
    
    return redirect_to(name, headers=make_cookie_headers(cookie))    

@view(r'^/(%s)/edit$' % WIKI_WORD)
def page_edit(environ, filler, name):

    conn = storage.connect()        
    page = storage.get_page(conn, name)

    if page: 
        content = page['content']
        page_title = page['title']
    else:
        content = u''
        content_hash = u''                        
        page_title = name

    message = u'' #make_message(get_message(environ))
    smartypants_enabled = USE_SMARTYPANTS and 'on' or 'off'

    author = get_author(environ)

    #page_title = u'Editing %s' % name
    
    return HTTP_OK, [], make_page('page_edit.html', filler, locals())
    
# -------------------
# Show page source
# -------------------

# @view(r'^/(%s)/source$' % WIKI_WORD)
# def page_source(environ, filler, name):
# 
#     conn = storage.connect()
#     page = storage.get_page(conn, name)
# 
#     # Sanity check
#     if not page:
#         raise NotFoundError
# 
#     return HTTP_OK, [('Content-Type', 'text/plain; charset=%s' % ENCODING)], page['content']

 
@view(r'^/(%s)/source/(\d+)$' % WIKI_WORD)
def page_source_revision(environ, filler, name, revision_id):

    conn = storage.connect()
    revision = storage.get_page_revision(conn, name, revision_id)

    # Sanity check
    if not revision:
        raise NotFoundError 

    return HTTP_OK, [('Content-Type', 'text/plain; charset=%s' % ENCODING)], revision['content']

# -------------------
# Show page revisions
# -------------------

@view(r'^/(%s)/revisions$' % WIKI_WORD)
def page_revisions(environ, filler, name):

    get_first = parse_query_string(environ)
    
    try:
        days = int(get_first('days', REVISION_LIFESPAN_DAYS))
    except ValueError:
        days = REVISION_LIFESPAN_DAYS

    conn = storage.connect()    
    revisions = storage.get_page_revisions(conn, name, days)
#@@TODO: sanity check     
# if not revisions:
#         raise NotFoundError(body=u'Page <i>%s</i> not found.' % (revision_id, name))

    
    #@@TODO:
    #page_title = revisions['title']
    page_title = name
                    
    template_dt = u'<tr><th colspan="4">$ddmmyyyy</th></tr>'
    template_dd = u'''<tr>
        <td class="dim"><a rel="nofollow" title="View source for revision $revision_id" href="$base_uri/$name/source/$revision_id">Rev. $revision_id</a> $diff_link</td>
        <td>$modified_on</td>
        <td class="author">$author_link from $host_name</td>
        <td class="revision-message">$revision_message</td>
        </tr>'''
        
    revision_count = 0
    revision_list = []    
    prev_ddmmyyyy = None
    for revision in revisions:  
    
        created_on = revision['created_on']
        revision_id = revision['revision_id']
        author = revision['author']
        ddmmyyyy = format_datetime(created_on, short=True)       
    
        filler_ = fill_template_namespace(
            name = name,
            revision_id = revision_id,
            host_name = revision['host_name'],
            author = revision['author'],
            revision_message = revision['message'],
        
            ddmmyyyy = ddmmyyyy,       
            modified_on = format_time(created_on), 
    
            diff_link = make_diff_link(name, revision_id),
            author_link = make_author_link(author)
        )
        
        if ddmmyyyy == prev_ddmmyyyy:                        
            revision_list.append(filler_(template_dd))
        else:
            revision_list.append(filler_(template_dt))
            revision_list.append(filler_(template_dd))

        prev_ddmmyyyy = ddmmyyyy        
        revision_count = revision_count + 1
    
    revisions=u'\n'.join(revision_list)
    s = (revision_count != 1) and u's' or u''

    if days:
        subhead = fill_template(u'Found $revision_count revision$s within the last $days days', **locals())
        listing_message = u'Older revisions may have been archived. <a href="$base_uri/$name/revisions?days=0">Show all revisions</a>.'
    else:        
        subhead = fill_template(u'Found $revision_count revision$s', **locals())
        listing_message = u''

    return HTTP_OK, [], make_page('page_revisions.html', filler, locals())

# -------------------
# Show revisions diff
# -------------------

@view(pattern=r'^/(%s)/diff/(\d+)/(\d+)$' % WIKI_WORD)
def page_diff(environ, filler, name, prev_revision_id, revision_id):

    conn = storage.connect()

    revision_a = storage.get_page_revision(conn, name, revision_id)
    revision_b = storage.get_page_revision(conn, name, prev_revision_id)

    # Sanity check
    if not revision_a:    
        raise NotFoundError  

    content = revision_a['content']  

    if revision_b:
        content_prev = revision_b['content']
    else:
        content_prev = u'' # Diff against empty page

    page_title = revision_a['title']

    from difflib import Differ

    content_prev, content = map(unicode.splitlines, [content_prev, content])

    diff = []
    for line in list(Differ().compare(content_prev, content)): 
        if line.startswith(u'+ '):
            diff.append(u'<ins>%s</ins><br />' % escape_html(line[2:]))
        elif line.startswith(u'- '):
            diff.append(u'<del>%s</del><br />' % escape_html(line[2:]))
        elif line.startswith(u'? '):
            pass
            #diff.append(u'<span>%s</span><br />' % escape_html(line[2:]))
        else:
            # As-is
            diff.append(u'%s<br />' % escape_html(line))
    
    diff = '\n'.join(diff)

    #page_title = u'Diff for %s' % name    
    
    if prev_revision_id:
        subhead = u'Showing differences between revision <b>%s</b> and <b>%s</b>' % (revision_id, prev_revision_id) 
    else:
        subhead = u'Page created'        
    
    return HTTP_OK, [], make_page('page_diff.html', filler, locals())



